//=============================================================================
// Motor Control Library
//
// Provides a servo motor functionality for NXT motors.
//
// Author: Gus Janson
//=============================================================================
// PID constants
const int MC_PID_msInterval = 40;

// Default PID constants
const int MC_KP = 100;
const int MC_KI = 10;
const int MC_KD = 300;

// Multiplier for PIC constants to allow for integer math
const int MC_KM = 100;

const long MC_propLimit = 100;
const int  MC_powerLimit = 100;
const int  MC_powerDiff = 10;

// Used in a encoder field that this entry is unused
const long MC_EUNUSED = 0x80000000;

// Number of integral sample stored
const int  MC_ISAMPLES = 10;

// A Base value used for motor proportion syncronization
// Gives a certain amount of minimum power and also avoids
// division by zero in math in case all syncronized motors
// are already at target position.
const int  MC_P_ABS_BASE = 5;

// Main public motor structure
typedef struct {
	long eTarget;     // encoder target (relaive to eZeroPos)
	long eZeroPos;    // actual encoder at user zero position
 	bool fSync;       // sync with other synced motors
	int KP;
	int KI;
	int KD;
	long eSlideTo;    // Used for target slide
} T_MCMotor;

T_MCMotor MCMotor[3];

// Internal flag that target is currently in a slide.
bool fSlide;

// Target slide structure unsed in MCLib to maintain data while
// target is sliding to a certain destination.
typedef struct {
	long eStart;    // encoder start position
	long eEnd;      // encoder end position
  long a;         // acceleration degrees / sec^2
  long v;         // velocity degrees / sec
  long tAtMaxV;   // t in ms when slide reaches max velocity
  long tAtBrake;  // t in ms when slide starts deacceleration
  long tAtEnd;    // t in ms when slide at eEnd
  long eAtMaxV;   // encoder when slide reaches max velocity
  long eAtBrake;  // encoder when slide starts deacceleration
} TSlideInfo;

TSlideInfo MotorSI[3];

bool UpdateSlideTarget(long t);

//=============================================================================
// Simple utility functions
long Limit(long l, long lLimit)
{
	if (l > lLimit)
	  return (lLimit);
	if (l < -lLimit)
	  return (-lLimit);
  return (l);
}

int max(int a, int b)
{
	if (a > b)
	  return a;
	else
	  return b;
}

int min(int a, int b)
{
	if (a < b)
	  return a;
	else
	  return b;
}

//=============================================================================
// Main Motor Control task
//
task _MCTask()
{
	int i;
	long eMotor[3];
	long eMotorPrev[3];
	long p[3], pl[3];
	long d[3];
	long iSample;
	int  power;
	int  powerPrev[3];
	int  powerNextMax, powerNextMin;
	int  iDataIndex;
	int iSum[3];                 // UNDONE Concsider using int here and all i data
	int iSumL;
	int iData[3][MC_ISAMPLES];
	long pAbs[3];
	long pAbsMax;
	long tSlide;


	// Init integral sample arrays
	for (i = 0; i < 3; i++) {
		iSum[i] = 0;
		for (iDataIndex = 0; iDataIndex < MC_ISAMPLES; iDataIndex++) {
			iData[i][iDataIndex] = 0;
		}
	}
  iDataIndex = 0;

  // Get initial motor encoder positions
	for (i = 0; i < 3; i++)
		eMotor[i] = nMotorEncoder[i] + MCMotor[i]. eZeroPos;

	// Init slide time to zero.  After slide this variable needs to be cleared for next
	// slide
	tSlide = 0;

	while (true) {
	  wait1Msec(MC_PID_msInterval);

	  pAbsMax = 0;

	  if (fSlide) {
	    fSlide = UpdateSlideTarget(tSlide);
	    if (fSlide)
	      tSlide += MC_PID_msInterval;
	    else
	      // Reset time variable for next slide
	      tSlide = 0;
	  }

	  for (i = 0; i < 3; i++) {
	  	if (MCMotor[i].eZeroPos != MC_EUNUSED) {
				eMotorPrev[i] = eMotor[i];
				eMotor[i] = nMotorEncoder[i] + MCMotor[i]. eZeroPos;

				p[i] = MCMotor[i].eTarget - eMotor[i];
				d[i] = eMotor[i] - eMotorPrev[i];

				pl[i] = Limit(p[i], MC_propLimit);

				if (MCMotor[i].fSync) {
					pAbs[i] = MC_P_ABS_BASE + abs(p[i]);
					if (pAbs[i] > pAbsMax)
					  pAbsMax = pAbs[i];
				}
	  	}
	  }

	  for (i = 0; i < 3; i++) {
	  	if (MCMotor[i].eZeroPos != MC_EUNUSED) {
  			iSample = pl[i];
				iSum[i] += iSample - iData[i][iDataIndex];
				iData[i][iDataIndex] = iSample;

				// Limit the iSum value used to no more then the limited proportion
				// value.  This is designed to remove overshoot caused by the i term.
				iSumL = pl[i]*MC_ISAMPLES;
				if (pl[i] > 0) {
				  if (iSum[i] < iSumL)
				    iSumL = iSum[i];
				} else {
				  if (iSum[i] > iSumL)
				    iSumL = iSum[i];
				}

				if (MCMotor[i].fSync && !fSlide) {
					pl[i] = pl[i] * pAbs[i] / pAbsMax;
			  }

        powerNextMax = max(20, powerPrev[i]+MC_powerDiff);
        powerNextMin = min(-20, powerPrev[i]-MC_powerDiff);

				power = (MCMotor[i].KP * pl[i] + MCMotor[i].KI * iSumL - MCMotor[i].KD * d[i])/ MC_KM;

				if (power > powerNextMax)
				  power = powerNextMax;

				if (power < powerNextMin)
				  power = powerNextMin;

				power = Limit(power, MC_powerLimit);

				motor[i] = power;

				powerPrev[i] = power;

				nxtDisplayTextLine(i+1, "%i %i", p[i], power);
      }
		}
		iDataIndex++; if (iDataIndex >= MC_ISAMPLES) iDataIndex = 0;
	}
}



void _MCInit(int nPort)
{
	motor[nPort] = 0;

  MCMotor[nPort].eTarget = 0;
  MCMotor[nPort].eZeroPos = 0;
  MCMotor[nPort].eSlideTo = MC_EUNUSED;

  MCMotor[nPort].fSync = false;

  // Set default P I D constants for this motor
  MCMotor[nPort].KP = MC_KP;
  MCMotor[nPort].KI = MC_KI;
  MCMotor[nPort].KD = MC_KD;
}

void MCInit(int nPort1, int nPort2 = -1, int nPort3 = -1)
{
	int iMotor;

	bFloatDuringInactiveMotorPWM = false;

	for (iMotor = 0; iMotor < 3; iMotor++) {
    MCMotor[iMotor].eZeroPos = MC_EUNUSED;
  }

	_MCInit(nPort1);
	if (nPort2 != -1)	_MCInit(nPort2);
	if (nPort3 != -1)	_MCInit(nPort3);
}

void MCSetSync(int nPort1, int nPort2 = -1, int nPort3 = -1)
{
	MCMotor[nPort1].fSync = true;
	if (nPort2 != -1)	MCMotor[nPort2].fSync = true;
	if (nPort3 != -1)	MCMotor[nPort3].fSync = true;
}

void MCFindLimit(int nPort, int nPower = 30, long eZeroPos = 0)
{
  int i;
  long e[3], eOld[3];
  int sign;
  bool fDone[3];
  int cNotDone;

  sign = sgn(nPower);

  // Init fDone to true as FindLimit will only apply to motors set to false
  fDone[0] = true;
  fDone[1] = true;
  fDone[2] = true;

  // If nPort is a synced motor, do the FindLimit on all synced motors
  if (MCMotor[nPort].fSync) {
  	cNotDone = 0;
    for (i=0; i<3; i++) {
    	if (MCMotor[i].fSync) {
				fDone[i] = false;
				cNotDone++;
    	}
    }
  } else {
  	fDone[nPort] = false;
  	cNotDone = 1;
  }

	// Get cur encoder and start motors
  for (i=0; i<3; i++) {
  	if (!fDone[i]) {
		  e[i] = nMotorEncoder[i];
			motor[i] = nPower;
  	}
  }

  // Give motors a little extra time to start moving
  wait1Msec(100);

  // Main FindLimit loop
	do {
	  wait1Msec(100);
    for (i=0; i<3; i++) {
    	if (!fDone[i]) {
		  	eOld[i] = e[i];
		    e[i] = nMotorEncoder[i];

		    // Is motor still making progress?
		    if (!(e[i]*sign > eOld[i]*sign)) {
		    	// No, reset and stop motor
		      fDone[i] = true;
		      cNotDone--;

				  nMotorEncoder[i] = 0;
					motor[i] = 0;

					MCMotor[i].eZeroPos = eZeroPos;
					MCMotor[i].eTarget = eZeroPos;
				}
		  }
		}
	} while(cNotDone > 0);
}

void MCStart()
{
	StartTask (_MCTask);
}

void MCSetTarget(int nPort, long eTarget)
{
  MCMotor[nPort].eTarget = eTarget;
}

void MCSetTargetR(int nPort, long eTargetR)
{
  MCMotor[nPort].eTarget += eTargetR;
}

#if 0
void MCUntilTarget(int nPort1, int nPort2 = -1, int nPort3 = -1)
{

}
#endif

//=============================================================================
// SlideTarget support
//
// Slide acceleration and max velocity
const long ST_A  = 600;   // degrees / sec^2
const long ST_V  = 300;   // degrees / sec

// Return true when slide is done
bool _UpdateSlideTarget(int nPort, long t)
{
	if (t < MotorSI[nPort].tAtMaxV) {
		// Accelerating - Note that t is in ms and a in deg/sec^2
		MCMotor[nPort].eTarget = MotorSI[nPort].eStart + MotorSI[nPort].a*t*t/2000000;
	} else if (t < MotorSI[nPort].tAtBrake){
		// At Velocity
		MCMotor[nPort].eTarget = MotorSI[nPort].eAtMaxV + (MotorSI[nPort].v*(t - MotorSI[nPort].tAtMaxV))/1000;
	} else if (t < MotorSI[nPort].tAtEnd) {
		// Breaking
		t = MotorSI[nPort].tAtEnd - t;   // Time from end for breaking
		MCMotor[nPort].eTarget = MotorSI[nPort].eEnd - MotorSI[nPort].a*t*t/2000000;
	} else {
		// Stop sliding
		MCMotor[nPort].eTarget = MotorSI[nPort].eEnd;
		return false;
	}
	return true;
}

bool UpdateSlideTarget(long t)
{
	int i;
	bool fNotDone;

	fNotDone = false;
	for (i=0; i<3; i++) {
		if (MCMotor[i].eSlideTo != MC_EUNUSED) {
			fNotDone |= _UpdateSlideTarget(i, t);
		}
	}
	return fNotDone;
}

void SlideTarget(long SlideV = ST_V, long SlideA = ST_A)
{
	int i;
  long degSlideDelta, degDist, degDistMax;
  long sign;
  long tA, tV;
  long degA, degV;
  long a;

nDatalogSize = 100;

  degDistMax = 0;
  for (i=0; i<3; i++) {
		if (MCMotor[i].eSlideTo != MC_EUNUSED) {
			degDist = abs(MCMotor[i].eSlideTo - MCMotor[i].eTarget);
			if (degDist > degDistMax) degDistMax = degDist;
		}
	}

	for (i=0; i<3; i++) {
		if (MCMotor[i].eSlideTo != MC_EUNUSED) {
		  // Set the start and end encoder positions
			MotorSI[i].eStart = MCMotor[i].eTarget;
			MotorSI[i].eEnd = MCMotor[i].eSlideTo;

			// Get slide delta (signed) and distance (positive)
			degSlideDelta = MotorSI[i].eEnd - MotorSI[i].eStart;
			degDist = abs(degSlideDelta);
			sign = sgn(degSlideDelta);

			// Scale the acceleration and velocity to the motor with max distance so that
			// this motor will arrive at the same time as the maxdist motor
			MotorSI[i].a = degDist * SlideA / degDistMax;
			MotorSI[i].v = degDist * SlideV / degDistMax;

			// Given acceleration and max velocity, calculate time and degrees needed
			// to accelerate from 0 to max velocity.
		  tA = (1000*MotorSI[i].v) / MotorSI[i].a;    // time in ms
		  degA = tA*MotorSI[i].v / 2000;        // tA in ms, SlideV in deg/seconds

			// degrees and time slide at max v
		  degV = degDist - degA*2;
		  tV = 1000*degV / MotorSI[i].v;

      // Determine if distance is enough to reach max velocity
			if (degDist > degA*2) {
				// Distance greater then needed for acceleration + break, will reach max velocity
				// Calculate
				MotorSI[i].eAtMaxV  = MotorSI[i].eStart + sign*degA;
				MotorSI[i].tAtMaxV  = tA;
				MotorSI[i].eAtBrake = MotorSI[i].eStart + sign*(degA+degV);
				MotorSI[i].tAtBrake = tA + tV;
				MotorSI[i].tAtEnd   = tA*2 + tV;
			} else {
				// Short distance, will not reach max velocity
				degA = degDist / 2;
//				tA = (int)(sqrt((float)degA * 2000000 / (float) MotorSI[i].a));
// RobotC bug!  The line above does not work but the one below does.
				a = MotorSI[i].a;
				tA = (long)(sqrt((2000000.0 * (float)degA) / (float) a));
				MotorSI[i].eAtMaxV  = MotorSI[i].eStart + sign*degA;
				MotorSI[i].tAtMaxV  = tA;
				MotorSI[i].eAtBrake = MotorSI[i].eStart + sign*degA;
				MotorSI[i].tAtBrake = tA;
				MotorSI[i].tAtEnd   = tA*2;
			}
			// Set the sign for a and v
			MotorSI[i].a *= sign;
			MotorSI[i].v *= sign;
		}
	}

SaveNxtDatalog();

  // Set slide flag so that _MCtask starts calling UpdateSlideTarget
  fSlide = true;

  // Wait for the _MCtask to finish and clear the flag
  while(fSlide)
    wait1Msec(50);
}
